
import Adafruit_BluefruitLE
from Adafruit_BluefruitLE.services import UART as OriginalUART
# from ble import uart
from ble.uart import UART
from ble.readdata import Device
import time
import atexit
import logging
log = logging.getLogger(__name__)

class App (object):
  """ A high level application object.

  Any application needing to talk to Dexcom G4 + Share will need
  to perform operations to setup the ble data transport.  This class
  mixes the UART, ble code, and provides helpful prolog and epilog
  routines that run before and after main, respectively.
  """
  def __init__ (self, **kwds):
    self.disconnect_on_after = kwds.get('disconnect_on_after', False)
    pass
  def setup_ble (self):
    self.remote = None
    self.ble = Adafruit_BluefruitLE.get_provider()
    # Initialize the BLE system.  MUST be called before other BLE calls!
    self.ble.initialize()
    # Get the first available BLE network adapter and make sure it's powered on.
    self.adapter = self.ble.get_default_adapter()
    self.adapter.power_on()
    log.info('Using adapter: {0}'.format(self.adapter.name))
    self.dexcom = None
    pass
  def setup_dexcom (self, serial=None, mac=None):
    # Once connected do everything else in a try/finally to make sure the device
    # is disconnected when done.
    try:
        # Wait for service discovery to complete for the UART service.  Will
        # time out after 60 seconds (specify timeout_sec parameter to override).
        # print device._device.GattServices
        log.info('Discovering services...')
        UART.discover(self.remote)

        # Once service discovery is complete create an instance of the service
        # and start interacting with it.
        self.uart = UART(self.remote, SERIAL=serial)


        self.dexcom = Device(self.uart)
        # log.info("DEXCOM", self.dexcom)
        if not self.dexcom:
          self.dexcom = Device(self.uart)
    except:
        # Make sure device is disconnected on exit.
        if self.disconnect_on_after:
          self.remote.disconnect()
  def prolog (self, clear_cached_data=True, disconnect_devices=True, scan_devices=True, connect=True, mac=None):
    """
    Things to do before running the main part of the application.
    """
    # Clear any cached data because both bluez and CoreBluetooth have issues with
    # caching data and it going stale.
    if clear_cached_data:
      self.ble.clear_cached_data()


    if disconnect_devices:
      # Disconnect any currently connected UART devices.  Good for cleaning up and
      # starting from a fresh state.
      log.info('Disconnecting any connected UART devices...')
      UART.disconnect_devices()

    if scan_devices:
      # Scan for UART devices.
      log.info('Searching for UART device...')
      try:
          if mac:
            self.remote = self.select_mac(mac=mac)
          else:
            self.adapter.start_scan()
            # Search for the first UART device found (will time out after 60 seconds
            # but you can specify an optional timeout_sec parameter to change it).
            self.remote = UART.find_device()
          if self.remote is None:
              raise RuntimeError('Failed to find UART device!')
      finally:
          # Make sure scanning is stopped before exiting.
          if self.adapter.is_scanning:
            self.adapter.stop_scan()

    if connect and not self.remote.is_connected:
      log.info('Connecting to device...')
      self.remote.connect()  # Will time out after 60 seconds, specify timeout_sec parameter
                        # to change the timeout.
    log.info(self.remote.name)
    # device._device.Pair( )
    # log.info(self.ble._print_tree( ))
    for service in self.remote.list_services( ):
      log.info("services: %s %s", service, service.uuid)
    log.info("ADVERTISED")
    log.info(self.remote.advertised)

    pass
  def select_mac (self, mac=None, **kwds):
    for device in self.enumerate_dexcoms(**kwds):
      if str(device.id) == mac:
        return device
  def enumerate_dexcoms (self, timeout_secs=10):
    self.adapter.start_scan()
    # Use atexit.register to call the adapter stop_scan function before quiting.
    # This is good practice for calling cleanup code in this main function as
    # a try/finally block might not be called since this is a background thread.
    def maybe_stop ( ):
      if self.adapter.is_scanning:
        self.adapter.stop_scan( )
    # atexit.register(maybe_stop)
    log.info('Searching for UART devices...')

    # print('Press Ctrl-C to quit (will take ~30 seconds on OSX).')
    # Enter a loop and print out whenever a new UART device is found.
    start = time.time( )
    now = time.time( )
    known_uarts = set()
    while (now - start) < timeout_secs:
        # Call UART.find_devices to get a list of any UART devices that
        # have been found.  This call will quickly return results and does
        # not wait for devices to appear.
        found = set(UART.find_devices())
        # Check for new devices that haven't been seen yet and print out
        # their name and ID (MAC address on Linux, GUID on OSX).
        new = found - known_uarts
        for device in new:
            log.info('Found UART: {0} [{1}]'.format(device.name, device.id))
        known_uarts.update(new)
        # Sleep for a second and see if new devices have appeared.
        time.sleep(1.0)
        now = time.time( )
    self.adapter.stop_scan( )
    return known_uarts

  def epilog (self):
    """
    Things to do after running the main part of the application.
    """
    # Make sure device is disconnected on exit.
    if self.disconnect_on_after and self.remote.is_connected:
      self.remote.disconnect()
    # self.ble._gobject_mainloop.quit( )
    pass
  def set_handler (self, handler):
    self.handler = handler
  def run (self):
    self.ble.run_mainloop_with(self.main)
    pass
  def main (self):
    """
    Subclasses should replace this method.
    """




import uuid

class Attrs:
    CradleService = uuid.UUID("F0ABA0B1-EBFA-F96F-28DA-076C35A521DB");

    # Share Characteristic Strings
    AuthenticationCode  = uuid.UUID("F0ABACAC-EBFA-F96F-28DA-076C35A521DB");
    ShareMessageReceiver = uuid.UUID("F0ABB20A-EBFA-F96F-28DA-076C35A521DB"); #  Max 20 Bytes - Writable
    ShareMessageResponse = uuid.UUID("F0ABB20B-EBFA-F96F-28DA-076C35A521DB"); #  Max 20 Bytes
    Command = uuid.UUID("F0ABB0CC-EBFA-F96F-28DA-076C35A521DB");
    Response = uuid.UUID("F0ABB0CD-EBFA-F96F-28DA-076C35A521DB"); #  Writable?
    HeartBeat = uuid.UUID("F0AB2B18-EBFA-F96F-28DA-076C35A521DB");

    # Possible new uuids????  60bfxxxx-60b0-4d4f-0000-000160c48d70
    CradleService2 = uuid.UUID("F0ACA0B1-EBFA-F96F-28DA-076C35A521DB");
    AuthenticationCode2  = uuid.UUID("F0ACACAC-EBFA-F96F-28DA-076C35A521DB"); #  read, write
    ShareMessageReceiver2 = uuid.UUID("F0ACB20A-EBFA-F96F-28DA-076C35A521DB"); #  read, write
    ShareMessageResponse2 = uuid.UUID("F0ACB20B-EBFA-F96F-28DA-076C35A521DB"); #  indicate, read
    Command2 = uuid.UUID("F0ACB0CC-EBFA-F96F-28DA-076C35A521DB"); #  read, write
    Response2 = uuid.UUID("F0ACB0CD-EBFA-F96F-28DA-076C35A521DB"); #  indicate, read, write
    HeartBeat2 = uuid.UUID("F0AC2B18-EBFA-F96F-28DA-076C35A521DB"); #  notify, read

    # Device Info
    DeviceService = uuid.UUID("00001804-0000-1000-8000-00805f9b34fb");
    PowerLevel = uuid.UUID("00002a07-0000-1000-8000-00805f9b34fb");

    VENDOR_UUID = uuid.UUID("F0ACA0B1-EBFA-F96F-28DA-076C35A521DB")





"""
openxshareble - pure python driver to communicate with Dexcom G4+Share
over ble.  Allows openaps to use ble to talk to Dexcom G4+Share.

Examples:
```
# add vendor to openaps
openaps vendor add openxshareble
# create device called share to use the new vendor
openaps device add share openxshareble
# use it, eg:
openaps use share iter_glucose 2
```

"""

import logging
log = logging.getLogger(__name__)
import os
import app
from openaps.uses.use import Use
from openaps.uses.registry import Registry


# we'll just replace all the original dexcom uses with new ones that swap
# out how the data is transported.
from openaps.vendors import dexcom

def configure_use_app (app, parser):
  pass

def configure_add_app (app, parser):
 pass

def configure_app (app, parser):
  pass

def configure_parser (parser):
  pass

def main (args, app):
  """
  """
  pass

def set_config (args, device):
  return ""
def display_device (device):
  return "/%s" % device.get('serial')

# get a new usage registry for this module
use = Registry( )
get_uses = use.get_uses

class BLEUsage (Use, app.App):
  """ Generic subclass for a single openaps use to use Dexcom's ble
  """

  __init__ = Use.__init__
  def before_main (self, args, app):
    """
    Get info such as serial from the device configuration.
    """

    logAddress = self.device.get('logAddress', '/dev/log')
    logLevel = getattr(logging, self.device.get('logLevel', 'INFO'))
    for handler in log.handlers[:]:
      log.removeHandler(handler)
    log.addHandler(logging.handlers.SysLogHandler(address=logAddress))
    log.setLevel(logLevel)
    serial = self.device.get('serial', None)
    mac = self.device.get('mac', None)
    # run prolog/setup
    self.prolog(mac=mac)
    # setup dexcom in particular, with configured serial number
    self.setup_dexcom(serial=serial)
  def after_main (self, args, app):
    # run any after code
    self.epilog( )
  def __call__ (self, args, app):
    """
    Using BLE requires dropping into glib's threading mechanics.
    """
    output = None
    # setup glib threading, configure a main loop in glib/dbus
    self.setup_ble( )
    def run ( ):
      """
      Run the main logic of the app, capturing the output.
      """
      self.before_main(args, app)
      output = self.main(args, app)
      self.after_main(args, app)
      return output
    # run the configured glib loop.
    res = self.ble.run_mainloop_with(run, quit_with_loop=False)
    return res

@use( )
class configure (Use):
  DEXCOMRX=os.environ.get('DEXCOMRX', '')
  __doc__ = """
  Configure DEXCOMRX serial number.


  Update the configuration so that your serial number is used to communicate
  with your Dexcom G4 with Share.
  Default: {DEXCOMRX}
  """.format(DEXCOMRX=DEXCOMRX)

  __init__ = Use.__init__
  def configure_app (self, app, parser):
    parser.add_argument('--serial', metavar='DEXCOMRX', default=os.environ.get('DEXCOMRX', ''), help="DexcomRX Serial Number")
    parser.add_argument('--mac', metavar='DEXCOMRX_MAC',
      default=os.environ.get('DEXCOMRX_MAC', ''),
      help="DexcomRX BLE MAC address")
  def get_params (self, args):
    conf = dict( )
    if args.serial.startswith('SM') and len(args.serial) > 8:
      conf.update(serial=args.serial)
    else:
      conf.update(serial=None)
    if len(args.mac.split(':')) == 6:
      conf.update(mac=args.mac)
    else:
      conf.update(mac=None)
    return conf
  def main (self, args, app):
    results = dict(serial=self.device.get('serial', None), mac=self.device.get('mac', None))
    params = self.get_params(args)
    dirty = False
    for field in [ 'serial', 'mac' ]:
      value = params.get(field)
      if getattr(self.device, 'extra', None):
        print "{field}={value}".format(field=field, value=results[field])
        log.info("{field}={value}".format(field=field, value=results[field]))
        if value:
          if value != results.get(field, None):
            dirty = True
            print "saving %s" % value
            results[field] = value
            self.device.extra.add_option(field, value)
    if dirty:
      self.device.store(app.config)
      app.config.save( )
    return results


@use( )
class list_dexcom (BLEUsage):
  """
  Scan the environment looking for Dexcom devices.
  
  Returns a list of scanned devices.
  """

  disconnect_on_after = False
  def before_main (self, args, app):
    pass
  def main (self, args, app):
    if not self.device.get('serial', None):
      self.disconnect_on_after = False
    receivers = self.enumerate_dexcoms( )
    results = [ ]
    for device in receivers:
      results.append(dict(name=str(device.name)
        , mac=str(device.id)
        , advertised=map(str, device.advertised)))
    return results

class DexcomTask (BLEUsage):
  """
  Generic utility to help swap original dexcom uses with new ones using
  the logic from this module to set up the usage.
  """
  disconnect_on_after = True
  @classmethod
  def Emulate (Klass, usage):
    """
    Given an original Use, an implementation, returns a transformed
    Use class so that our logic controls how data gets transported.
    """
    class EmulatedUsage (usage, Klass):
      __doc__ = usage.__doc__
      __name__ = usage.__name__

      before_main = Klass.before_main
      after_main = Klass.after_main
      __call__ = Klass.__call__

    EmulatedUsage.__name__ = usage.__name__
    return EmulatedUsage

def transform (existing, Emulator, Original):
  """
  helper method to extract substitutes from original list of uses.
  """
  results = dict( )
  for name, usage in existing.use.__USES__.items( ):
    if issubclass(usage, Original):
      adapted = Emulator.Emulate(usage)
      adapted.__name__ = name
      if name not in results:
        results[name] = adapted
  return results


adapted = transform(dexcom, DexcomTask, dexcom.scan)
use.__USES__.update(**adapted)





from dexcom_reader import readdata
import xml.etree.ElementTree as etree


class Device (readdata.Dexcom):
  """
  Local shim of original readdata.Dexcom from dexcom_reader.
  This simply replaces some of the logic to set up the data transport
  mechanisms.
  """
  def __init__ (self, uart):
    # assign serial port/uart device
    self.uart = uart

  def Connect (self):
    return True

  @property
  def port (self):
    return self.uart

  def Disconnect (self):
    return True

  def flush (self):
    return True

  def write (self, data, *args, **kwargs):
    # the usb protocol is exactly the same, the ble protocol adds a
    # prefix of 0x01, 0x01 to every message written.
    prefix = str(bytearray([ 0x01, 0x01 ]))
    return self.port.write(prefix + data, *args, **kwargs)





import Adafruit_BluefruitLE
from Adafruit_BluefruitLE.services import UART as OriginalUART
import Queue
import uuid
import time
from attrs import Attrs
import logging
log = logging.getLogger(__name__)

class ShareUART (OriginalUART):
  ADVERTISED = [Attrs.CradleService]
  # SERVICES = [Attrs.DeviceService]
  SERVICES = [Attrs.CradleService]
  CHARACTERISTICS = [Attrs.AuthenticationCode, Attrs.Command, Attrs.Response, Attrs.ShareMessageReceiver, Attrs.ShareMessageResponse, Attrs.HeartBeat, Attrs.DeviceService, Attrs.PowerLevel]

  UART_SERVICE_UUID = Attrs.CradleService
  TX_CHAR_UUID = Attrs.Command
  RX_CHAR_UUID = Attrs.Response
  pass
class Share2UART (OriginalUART):
  # ADVERTISED = [Attrs.CradleService2]
  # ADVERTISED = [Attrs.VENDOR_UUID]
  ADVERTISED = [Attrs.VENDOR_UUID]
  # SERVICES = [Attrs.DeviceService]
  # SERVICES = [Attrs.CradleService2, Attrs.VENDOR_UUID]
  SERVICES = [Attrs.VENDOR_UUID, Attrs.DeviceService]
  # CHARACTERISTICS = [Attrs.AuthenticationCode2, Attrs.Command2, Attrs.Response2, Attrs.ShareMessageReceiver2, Attrs.ShareMessageResponse2, Attrs.HeartBeat2, Attrs.DeviceService, Attrs.PowerLevel]
  CHARACTERISTICS = [ ]

  HEARTBEAT_UUID = Attrs.HeartBeat2
  # UART_SERVICE_UUID = Attrs.CradleService2
  UART_SERVICE_UUID = Attrs.VENDOR_UUID
  TX_CHAR_UUID = Attrs.ShareMessageReceiver2
  RX_CHAR_UUID = Attrs.ShareMessageResponse2
  SendDataUUID = Attrs.ShareMessageReceiver2
  RcveDataUUID = Attrs.ShareMessageResponse2
  CommandUUID  = Attrs.Command2
  ResponseUUID = Attrs.Response2
  AUTH_UUID    = Attrs.AuthenticationCode2
  def __init__(self, device, **kwds):
      """Initialize UART from provided bluez device."""
      # Find the UART service and characteristics associated with the device.
      log = logging.getLogger(__name__)
      self.log = log.getChild('uart')
      self._uart = device.find_service(self.UART_SERVICE_UUID)
      log.info("UART %s", self._uart)
      self._queue = Queue.Queue()
      r = device.is_paired
      self.serial = kwds.pop('SERIAL', None)
      log.info("paired? %s", r)
      if not r:
        log.info("pairing...")
        # help(device._device)
        # help(device._device.Pair)
        device.pair( )
        # device._device.Pair( )
        log.info("paired")
        log.info(device.advertised)
        log.info("finding service")
        self._uart = device.find_service(self.UART_SERVICE_UUID)
        log.info("SERVICE %s", self._uart)
        self.pair_auth_code(self.serial)
      """
      for svc in device.list_services( ):
        log.info("{0}, {1}, {2}, {3}", svc.uuid, svc.uuid == self.UART_SERVICE_UUID, svc, svc._service)
        log.info("CHARACTERISTICS")
        chrsts = svc.list_characteristics( )
        for chtr in chrsts:
          log.info("{0} {1} {2}", chtr.uuid, chtr, chtr._characteristic)
      """
      # print device.list_services( )
      self.setup_dexcom( )
  def set_serial (self, SERIAL):
    self.serial = SERIAL
  def pair_auth_code (self, serial):
      self.log.info("sending auth code %s", serial)
      self._auth = self._uart.find_characteristic(self.AUTH_UUID)
      self.log.info(self._auth)
      # self._auth.
      msg = bytearray(serial + "000000")
      self._auth.write_value(str(msg))
  def setup_dexcom_heartbeat (self):
    self._heartbeat = self._uart.find_characteristic(self.HEARTBEAT_UUID)
  def do_heartbeat (self):
    if not self._heartbeat.notifying:
      self._heartbeat.start_notify(self._heartbeat_tick)
  def setup_dexcom (self):
    self.remainder = bytearray( )
    self._tx = self._uart.find_characteristic(self.TX_CHAR_UUID)
    self._rx = self._uart.find_characteristic(self.RX_CHAR_UUID)
    # Use a queue to pass data received from the RX property change back to
    # the main thread in a thread-safe way.
    self.setup_dexcom_heartbeat( )
    self.do_heartbeat( )
    """
    self._heartbeat = self._uart.find_characteristic(self.HEARTBEAT_UUID)
    if not self._heartbeat.notifying:
      self._heartbeat.start_notify(self._heartbeat_tick)
    """
    self._char_rcv_data = self._uart.find_characteristic(self.RcveDataUUID)
    if self._rx.notifying:
      self._rx.stop_notify( )
    if not self._rx.notifying:
      self._rx.start_notify(self._rx_received)
  def _heartbeat_tick (self, data):
    self.log.info("_heartbeat_tick %s", str(data).encode('hex'))
  def _on_rcv (self, data):
    self.log.info("_on_rcv %s", str(data).encode('hex'))

  def read (self, size=1, timeout_sec=None):
    spool = bytearray( )
    spool.extend(self.remainder)
    self.remainder = bytearray( )
    while len(spool) < size:
      spool.extend(self.pop(timeout_sec=timeout_sec))
      time.sleep(.100)
    self.remainder.extend(spool[size:])
    return str(spool[:size])
  def pop (self, timeout_sec=None):
    return super(Share2UART, self).read(timeout_sec=timeout_sec)

class BothShare (ShareUART):
  ADVERTISED = ShareUART.ADVERTISED + Share2UART.ADVERTISED
  # SERVICES = [Attrs.DeviceService]
  SERVICES =  ShareUART.SERVICES + Share2UART.SERVICES
  CHARACTERISTICS =  ShareUART.SERVICES + Share2UART.SERVICES
  

  UART_SERVICE_UUID = Attrs.CradleService2
  TX_CHAR_UUID = Attrs.Command2
  RX_CHAR_UUID = Attrs.Response2
  pass

class UART (Share2UART):
  pass




# from ez_setup import use_setuptools
# use_setuptools()
from setuptools import setup, find_packages

setup(name              = 'openxshareble',
      version           = '0.0.0',
      author            = 'Ben West',
      author_email      = 'bewest@gmail.com',
      description       = 'Python library for interacting with Dexcom G4 with Share over Bluetooth low energy.',
      license           = 'MIT',
      url               = 'https://github.com/bewest/openxshareble/',
      packages          = find_packages())
